{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE PackageImports #-}
{-# LANGUAGE RecordWildCards #-}

module TiedoteMD.Read where

import Control.Exception (throwIO, catch)
import Control.Monad ((>=>), join, unless)
import Control.Monad.IO.Class (liftIO, MonadIO)
import Control.Monad.Reader (ReaderT, MonadReader, runReaderT, asks)
import Control.Monad.Except (MonadError, throwError)
import Control.Monad.Trans (lift)
import Crypto.Hash (hashWith, SHA256(..))
import Data.Acid (AcidState, update)
import Data.Bifunctor (first, second)
import Data.ByteArray (convert)
import Data.ByteString (ByteString)
import "base64" Data.ByteString.Base64.URL
import Data.Default (def)
import Data.Either (rights, lefts, fromRight)
import Data.FileStore (FileStore(..), Revision(..), FileStoreError(..), gitFileStore)
import Data.List (singleton, isSuffixOf)
import Data.MIME (ContentID, makeContentID, renderContentID)
import Data.Maybe (fromMaybe)
import Data.Text.Encoding (decodeUtf8, encodeUtf8)
import Data.Time (UTCTime, zonedTimeToUTC)
import Data.Time (getCurrentTime)
import Data.Time.Format.ISO8601 (iso8601ParseM)
import System.Exit (ExitCode(..))
import System.FilePath ((</>), takeDirectory, normalise)
import System.FilePath.Glob (match, compile)
import System.IO (hClose)
import System.Process (createProcess, proc, waitForProcess, CreateProcess(..), StdStream(..))
import Text.Pandoc (Pandoc(..), PandocMonad, PandocIO, PandocError(..))
import Text.Pandoc.MediaBag (MediaItem(..), lookupMedia, mediaItems)
import Text.Pandoc.Readers (readMarkdown)
import Text.Pandoc.Walk (walkM)
import Text.Pandoc.Writers (writePlain, writeMarkdown, writeHtml5String)

import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as LBS
import qualified Data.Map as Map
import qualified Data.Set as Set
import qualified Data.Text as T
import qualified Data.Text.IO as T
import qualified Text.Pandoc as Pandoc

import TiedoteMD.Git
import TiedoteMD.State
import TiedoteMD.Templates
import TiedoteMD.Types

updateMessages :: AcidState State -> FilePath -> IO ()
updateMessages acid repoPath =
    gitPull repoPath >> readMessageFiles (gitFileStore repoPath) >>= update acid . UpdateState

readMessageFiles :: FileStore -> IO [Message]
readMessageFiles store = do
    files <- index store 
    messages <- mapM (readMessageFile store) $ filter (".md" `isSuffixOf`) files
    mapM_ (T.putStrLn . renderError) $ lefts messages
    currentTime <- getCurrentTime
    print currentTime
    pure $ filter ((currentTime <) . sendTime) $ rights messages
-- TODO: files could be read in parallel instead of sequentially with mapM

newtype ReadM a = ReadM (ReaderT ReadMState PandocIO a)
    deriving (Functor, Applicative, Monad, MonadIO, MonadReader ReadMState, MonadError PandocError)

data ReadMState = ReadMState
    { fileStoreHandle :: FileStore
    , currentFile :: FilePath
    }

liftPandoc :: PandocIO a -> ReadM a
liftPandoc = ReadM . lift

runReadM :: FileStore -> FilePath -> ReadM a -> IO (Either PandocError a)
runReadM store file (ReadM m) = Pandoc.runIO $ runReaderT m $ ReadMState store file

readStoreFile :: FilePath -> ReadM LBS.ByteString
readStoreFile file = do
    store <- asks fileStoreHandle
    errorOrFile <- liftIO $ (Right <$> retrieve store file Nothing)
        `catch` \NotFound -> pure $ Left $ PandocResourceNotFound $ T.pack file
    either throwError pure errorOrFile

getCurrentDirectory :: ReadM FilePath
getCurrentDirectory = takeDirectory <$> asks currentFile

getRealPath :: FilePath -> ReadM FilePath
getRealPath file = normalise . (</> file) <$> getCurrentDirectory

instance PandocMonad ReadM where
    readFileLazy = getRealPath >=> readStoreFile
    readFileStrict = getRealPath >=> fmap LBS.toStrict . readStoreFile
    getDataFileName = pure
    getModificationTime file = do
        store <- asks fileStoreHandle
        file' <- getRealPath file
        liftIO $ fmap revDateTime . revision store =<< latest store file'
    fileExists file = do
        store <- asks fileStoreHandle
        file' <- getRealPath file
        liftIO $ (const True <$> latest store file') `catch` \NotFound -> pure False
    glob pattern = do
        store <- asks fileStoreHandle
        directory <- getCurrentDirectory
        filter (match $ compile $ directory </> pattern) <$> liftIO (index store)
    lookupEnv = liftPandoc . Pandoc.lookupEnv
    getCurrentTime = liftPandoc Pandoc.getCurrentTime
    getCurrentTimeZone = liftPandoc Pandoc.getCurrentTimeZone
    newStdGen = liftPandoc Pandoc.newStdGen
    newUniqueHash = liftPandoc Pandoc.newUniqueHash
    openURL = liftPandoc . Pandoc.openURL
    readStdinStrict = pure mempty
    getCommonState = liftPandoc Pandoc.getCommonState
    putCommonState = liftPandoc . Pandoc.putCommonState
    logOutput = liftPandoc . Pandoc.logOutput

readMessageFile :: FileStore -> FilePath -> IO (Either Error Message)
readMessageFile store file = retrieve store file Nothing >>=
    uncurry (parseMessageFile store file) . addHash . T.filter (/= '\r') . decodeUtf8 . BS.toStrict
        where addHash text = (convert $ hashWith SHA256 $ encodeUtf8 text, text)

cidOf :: MediaItem -> ContentID
cidOf MediaItem {mediaContents, mediaMimeType} =
    fromRight (error "makeContentID failed with valid input!") $
        makeContentID $ "<" <> encodedHash <> "@tiedote.md.sha256>"
            where encodedHash = encodeBase64' $ convert $ hashWith SHA256 $
                      LBS.toStrict mediaContents <> encodeUtf8 mediaMimeType

cidUrl :: ContentID -> T.Text
cidUrl = mappend "cid:" . stripAngleBrackets . decodeUtf8 . renderContentID
    where stripAngleBrackets = T.init . T.tail
          -- The string is always surrounded by < > so this is safe

makeMediaPart :: MediaItem -> MediaPart
makeMediaPart mediaItem@MediaItem {..} = MediaPart
    { mediaPartMimeType = encodeUtf8 mediaMimeType
    , mediaPartContentID = renderContentID $ cidOf mediaItem
    , mediaPartContents = LBS.toStrict mediaContents
    }

reconstructMediaItem :: (FilePath, T.Text, LBS.ByteString) -> MediaItem
reconstructMediaItem (mediaPath, mediaMimeType, mediaContents) = MediaItem {..}

replaceImagesWithCid :: PandocMonad m => Pandoc -> m Pandoc
replaceImagesWithCid = Pandoc.fillMediaBag >=> walkM handleImage
    where handleImage (Pandoc.Image attr lab (src, title)) = do
              mediaItem <- (\media -> fromMaybe (error $ "fillMediaBag left an image uncollected! impossible!")
                           $ lookupMedia (T.unpack src) media) <$> Pandoc.getMediaBag
              pure $ Pandoc.Image attr lab (cidUrl $ cidOf mediaItem, title)
          handleImage x = pure x

parseMessageFile :: FileStore -> FilePath -> ByteString -> T.Text -> IO (Either Error Message)
parseMessageFile store file hash text = fmap (join . first PandocError) . runReadM store file $ do
    pandoc@(Pandoc meta _) <- flip readMarkdown text
        def {Pandoc.readerStandalone = True, Pandoc.readerExtensions = Pandoc.pandocExtensions} >>= replaceImagesWithCid
    let tiedoteMeta = do
            previewTo <- lookupMeta' "tarkistaja" meta >>= metaToEmails
            previewTime <- lookupMeta' "deadline" meta >>= metaToTime
            sendTime <- lookupMeta' "lähetysaika" meta >>= metaToTime
            subject <- lookupMeta' "otsikko" meta >>= metaToText
            pure (previewTo, previewTime, sendTime, subject)
    case tiedoteMeta of
      (Left err) -> pure $ Left err
      Right (previewTo, previewTime, sendTime, subject) -> do
          plainTextMessage <- renderHelper writePlain plainTemplate pandoc
          markdownMessage <- renderHelper writeMarkdown markdownTemplate pandoc
          let htmlPandoc = let Pandoc meta' blocks' = pandoc in Pandoc
                  (Pandoc.Meta $ Map.insertWith (flip const) "pagetitle" (Pandoc.MetaString subject) $ Pandoc.unMeta meta')
                  blocks'
          htmlMessage <- liftIO . inlineCSS =<< renderHelper writeHtml5String htmlTemplate htmlPandoc
          mediaParts <- Set.fromList . map (makeMediaPart . reconstructMediaItem) . mediaItems <$> Pandoc.getMediaBag
          pure $ pure $ Message
            { recipients = []
            , messageHash = hash
            , messageContent = MessageContent {..}
            , previewMailID = Nothing
            , ..
            }
        where renderHelper writer template = writer (writerOpts {Pandoc.writerTemplate = Just template})
              writerOpts = def
                  { Pandoc.writerTOCDepth = 2
                  , Pandoc.writerTableOfContents = True
                  , Pandoc.writerSectionDivs = True
                  }

inlineCSS :: T.Text -> IO T.Text
inlineCSS html = do
    (Just hIn, Just hOut, _, processHandle) <- createProcess $
        (proc "douceur" ["inline", "/dev/stdin"]) {std_in = CreatePipe, std_out = CreatePipe}
    T.hPutStrLn hIn html
    hClose hIn
    inlined <- T.hGetContents hOut
    exitCode <- waitForProcess processHandle
    unless (exitCode == ExitSuccess) $ throwIO $ ProcessError "douceur" exitCode
    pure inlined

metaToTime :: Pandoc.MetaValue -> Either Error UTCTime
metaToTime meta = do
    textMeta <- metaToTexts meta >>= headOrError
    maybe (Left $ InvalidTime textMeta) pure $
        fmap zonedTimeToUTC $ iso8601ParseM $ T.unpack $ T.strip textMeta
            where headOrError (x:_) = pure x
                  headOrError _ = Left $ InvalidTime ""

metaToEmails :: Pandoc.MetaValue -> Either Error [Email]
metaToEmails meta = metaToTexts meta >>= mapM
    (\text -> maybe (Left $ InvalidEmail text) pure $ email $ encodeUtf8 $ T.strip text)

metaToText :: Pandoc.MetaValue -> Either Error T.Text
metaToText = second T.unwords . metaToTexts

metaToTexts :: Pandoc.MetaValue -> Either Error [T.Text]
metaToTexts (Pandoc.MetaString text) = pure [text]
metaToTexts (Pandoc.MetaInlines inlines) = second singleton $ inlinesToText inlines
metaToTexts (Pandoc.MetaList metas) = second concat $ mapM metaToTexts metas
metaToTexts _ = Left InvalidData

inlinesToText :: [Pandoc.Inline] -> Either Error T.Text
inlinesToText = first PandocError . Pandoc.runPure . writeMarkdown def{Pandoc.writerExtensions = Pandoc.pandocExtensions} . Pandoc Pandoc.nullMeta . singleton . Pandoc.Plain

lookupMeta' :: T.Text -> Pandoc.Meta -> Either Error Pandoc.MetaValue
lookupMeta' key meta = maybe (Left $ NoMeta key) Right $ Pandoc.lookupMeta key meta
