{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RecordWildCards #-}

module TiedoteMD.Review where

import Control.Concurrent (threadDelay)
import Control.Exception (throwIO)
import Control.Lens.Combinators (view, filtered, toListOf, _Right, set, over)
import Control.Monad (forever, when)
import Data.Acid (AcidState, update, query)
import Data.Bifunctor (first)
import Data.Binary.Builder (toLazyByteString)
import Data.ByteString (ByteString)
import Data.IMF (body)
import Data.List (sortBy)
import Data.MIME (Address(..), MIMEMessage, MIME(..), Mailbox(..), ReplySettings, ReplyMode(..), mime, message, parse, matchContentType, contentType, entities, transferDecoded', headerInReplyTo, buildMessage, reply, defaultReplySettings, replyMode, headerCC, headerTo, headerBCC, headerReferences)
import Data.MIME.Charset (charsetText', defaultCharsets)
import Data.Maybe (catMaybes)
import Data.Maybe (listToMaybe, fromMaybe)
import Data.Ord (comparing)
import Data.Text.Encoding (encodeUtf8, decodeUtf8)
import Data.Time (getCurrentTime, diffUTCTime, nominalDay, secondsToNominalDiffTime)
import Data.Time.Format (formatTime, defaultTimeLocale)
import System.Random (getStdRandom, uniform)

import qualified Data.ByteString as BS
import qualified Data.List.NonEmpty as NE
import qualified Data.Text as T

import TiedoteMD.Send (renderMessage', sendmail)

import TiedoteMD.State
import TiedoteMD.Types

manageIncomingEmail :: AcidState State -> Mailbox -> FilePath -> IO ()
manageIncomingEmail acid sender sendmailPath = do
    mail <- BS.getContents >>= either throwIO pure . parseEmail 
    let inReplyTo = concat $ map (map messageIDToMailID) $
            toListOf headerInReplyTo mail <> toListOf headerReferences mail
            -- References should contain In-Reply-To, but I wouldn't trust every
            -- mailer to include References
        recipients = getEmailsFromMessage mail
    updatedMessages <- update acid $ AddMessageRecipients inReplyTo recipients
    mapM_ (sendRecipientVerifications mail sender sendmailPath) updatedMessages
    when (null updatedMessages) $ sendmail sendmailPath $ toLazyByteString $ buildMessage $
        set body (Part $ encodeUtf8 $ "Vastaavaa tiedote-esikatselua ei löytynyt. Tarkista, että vastasit tiedotteen uusimpaan esikatseluun ja ettei tiedotteen lähetysaika ole jo umpeutunut.\n\nMikäli tämä ei ratkaise ongelmaa, ole yhteydessä järjestelmän ylläpitäjään.") $
        reply defaultCharsets (replySettings sender) mail

managePreviews :: AcidState State -> Mailbox -> FilePath -> IO ()
managePreviews acid sender sendmailPath = forever $ do
    maybeMessage <- getCurrentTime >>= query acid . GetMessageToPreview
    case maybeMessage of 
      Nothing -> pure ()
      Just msg@(Message {..}) -> do
          mailID <- uniqueMailID sender
          boundary <- getStdRandom uniform
          boundary' <- getStdRandom uniform
          let mail = renderMessage' "Esikatselu: " (Just mailID) msg sender boundary boundary'
          sendmail sendmailPath $ toLazyByteString $ buildMessage $
              set (headerTo defaultCharsets) (map (Single . emailToMailbox) previewTo) mail
          update acid $ SetPreviewID messageHash mailID
    threadDelay $ 1000 * 1000

sendRecipientVerifications :: MIMEMessage -> Mailbox -> FilePath -> Message -> IO ()
sendRecipientVerifications mail sender sendmailPath Message {..} = do
    currentTime <- getCurrentTime
    let recipientCount = T.pack $ show $ length recipients
        formattedSendTime = T.pack $ formatTime defaultTimeLocale "%d.%m.%Y klo %H.%M (UTC%Ez)" sendTime
        formatTimeDiff timediff
          | timediff > nominalDay = formatTime defaultTimeLocale
              "%d päivän, %H tunnin ja %M minuutin päästä" timediff
          | timediff > secondsToNominalDiffTime 3600 = formatTime defaultTimeLocale
              "%h tunnin ja %M minuutin päästä" timediff
          | timediff > secondsToNominalDiffTime (10 * 60) = formatTime defaultTimeLocale
              "%m minuutin päästä" timediff
          | timediff > secondsToNominalDiffTime 60 = formatTime defaultTimeLocale
              "%m minuutin ja %S sekunnin päästä" timediff
          | otherwise = formatTime defaultTimeLocale "%s sekunnin päästä" timediff
        sendingAfter = T.pack $ formatTimeDiff $ diffUTCTime sendTime currentTime
    sendmail sendmailPath $
        toLazyByteString $ buildMessage $
        setHeaderCC previewTo $
        set body (Part $ encodeUtf8 $
            (("Viesti lähetetään seuraaviin " <> recipientCount <> " osoitteiseen " <> formattedSendTime <>
              ", eli " <> sendingAfter <> ". Lähetä vastauksena uusi lista, mikäli tämä on" <>
              " virheellinen.\n\n") <>) $
                T.intercalate "\n" $ map (decodeUtf8 . emailToByteString) recipients) $
        reply defaultCharsets (replySettings sender) mail
        where addressToMailboxes (Single mb) = [mb]
              addressToMailboxes (Group _ mbs) = mbs
              sameRecipient (Mailbox _ addr1) (Mailbox _ addr2) = addr1 == addr2
              setHeaderCC reviewers mail' = 
                  let to = concatMap addressToMailboxes $ view (headerTo defaultCharsets) mail'
                      cc = concatMap addressToMailboxes $ view (headerCC defaultCharsets) mail'
                      bcc = concatMap addressToMailboxes $ view (headerBCC defaultCharsets) mail'
                      -- XXX: bcc is only for completeness; there should never be any BCC recipients
                      ccsToAdd = filter (not . flip (elemBy sameRecipient) (to <> cc <> bcc)) $
                          map emailToMailbox reviewers
                  in over (headerCC defaultCharsets) (<> map Single ccsToAdd) mail'

parseEmail :: ByteString -> Either Error MIMEMessage
parseEmail = first (EmailParseError . T.pack) . parse (message mime)

getEmailsFromMessage :: MIMEMessage -> [Email]
getEmailsFromMessage = readEmails . toListOf
    (entities . filtered isText . transferDecoded' . _Right . charsetText' defaultCharsets . _Right)
    where isText = matchContentType "text" Nothing . view contentType

readEmails :: [T.Text] -> [Email]
readEmails = fromMaybe [] . listToMaybe . reverse . sortBy (comparing length) .
    map (catMaybes . map (email . encodeUtf8) . T.lines)

replySettings :: Mailbox -> ReplySettings
replySettings sender = set replyMode ReplyToGroup $ defaultReplySettings $ NE.singleton sender
